// TC_SERV.C - Linux parallel port server for Total Commander
// (Transfer rate 60 KB/sec or better)

// Disclaimer:  TC_SERV is passed on in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// (TC is a file manager, and as such its reason of existence is the
//  handling and backing up of precious data files.  However, TC may
//  one day fail in the least opportune way.  You have been warned!
//  Always use multiple and diverse backup strategies.)
//
// Original code, in Pascal,	   Copyright 2000-2005 Christian Ghisler.
// Transferred to C, and extended, Copyright 2004-2005 Paul Theo Pilgram.
//
// Compile using one of the 'compilefor_...' scripts:
//   chmod  +x  compilefor_UTF8.sh
//   compilefor_UTF8.sh
// (You will be asked for the root password)
//
// Or compile/install manually:
//   gcc -o tc_serv  tc_serv.c
//   su
//   chown root tc_serv
//   chmod  +s  tc_serv
//   exit
//
// Run with:
//   ./tc_serv
// or:
//     tc_serv		 (after you have moved it by:   mv  tc_serv  /bin  )
//
// The program will call  seteuid(getuid())  right after claiming the port .

// Normally, TC_SERV is called without parameters.
// It tries then to connect via any parallel port, taking whichever
// is the first to receive data (it's usually the port at 0x378).
// Alternatively, you can specify the port number: 1, 2, or 3
// (e.g. in case you have a printer on another parallel port).

// Remember: parallel interfaces are NOT HOT PLUGGABLE.
// Connect and disconnect the cable only with both computers switched off.

// This server
//  o  supports (a good part of) the Total Commander functionality
//     on a remote computer, especially if it is a Linux machine,
//  o  handles character set conversion of file/dir names and of text
//     files (notably UTF8, the Euro symbol, and MS-DOS ASCII),
//  o  is ready (though untested) for computers with different byte order,
//  o  can serve as a blueprint for further servers.

// Tested on    SuSE 10.0 i386 Linux (gcc version 4.0.2).
// Tested on    SuSE  9.1 i386 Linux (gcc version 3.3.3).
// Tested on    SuSE  9.0 i386 Linux (gcc version 3.3.1).
// Tested on RedHat FedC3 i386 Linux (gcc version 3.4.2).
// (The program is UNTESTED on DOS, i.e. for running like 'LPTDOS'.
//  It cannot run on Windows due to lack of port access.)
// So far it has been tested only on little-endian CPUs, namely the
// Intel i386 architecture.  Provisions have been made so it runs on
// big-endian CPUs (e.g. Suns, Motorola) just as well, but there was no
// chance for verifying this.  We would be interested in reports on how
// TC_SERV runs on such CPUs.  (Do such machines have a parallel port?)

// This program was written on a setup of Windows and Linux computers
// with a facility for transferring files only from Windows to Linux.
// It was hence written in a weird C dialect mix which suits both
// Microsoft C and Gnu C.  After all, TC_SERV was not available yet...

#include  <stdio.h>

// #define USING_UTF8		// (e.g. SuSE >9.0) Gets set by the script

#define  true	1
#define false	0

#if !defined(_MSC_VER)		// Microsoft C compiler --> not Linux
#define LINUX	1
#endif


// #if defined(_MSC_VER)// stdafx.h is required by Microsoft Visual C, but
// #include "stdafx.h"	// Visual C hates seeing this #ifdef-ed in (-- why?)
// #endif


#include    <stdlib.h>		// [W]_sleep   [W]errno_decl
#include    <string.h>
#include     <errno.h>		// [W]errno_constants  [U]errno_decl
 // The doserror() codes are: ...   see <errno.h> on Windows
 // Amazing: Same codes Windows=Linux: /usr/include/asm-generic/errno-base.h
 // Variable/function 'errno' has been eliminated.

#include    <signal.h>		// [U]signal
#include <sys/types.h>
#include  <sys/stat.h>		// _stat  _fstat  S_IFDIR  [U]mkdir
// [U]: bits/stat.h defines struct stat but MUST be included only via sys/stat.h
				// The leading underscores work even in Linux
#include      <time.h>		// localtime  mktime
#include <sys/timeb.h>		// ftime


static   struct stat	fprops ;// Parameter/result block for 'stat' + 'fstat'
static   struct timeb	time_a, time_b;
#define  GetEpochTime()	(ftime(&time_b), time_b.time - time_a.time) // Timer


// Time/date stamps:

static union {
  struct {	// bit fields
    unsigned int  secs : 5	;//   0...31	       (seconds*31+29)/59
    unsigned int  mins : 6	;//   0...59 (...63)    minutes
    unsigned int  hrs  : 5	;//   0...23 (...31)    hours
    unsigned int  day  : 5	;// 0,1...31		day
    unsigned int  month: 4	;// 0,1...12 (...15)    month
    unsigned int  year : 7	;//   0...127		years after 1980
// -- In MSVC and Gnu/Intel, bits go from right to left -- any guarantee ???
  }    f;
  long L;
}		 fat_time	;// FAT32 time stamp (16+16 bits)

static struct tm  ux_time	;// Unix  time stamp

static int	  uid = 999	;// User UID (who am I)


#if defined(LINUX)

#include <endian.h>		// __BYTE_ORDER __LITTLE_ENDIAN __BIG_ENDIAN
#include <sys/io.h>		// inb  outb  ioperm
#include  <utime.h>		// utime
#include <unistd.h>		// sleep[seconds!]    usleep[usecs]
#define _sleep(a)    usleep((a)*1000L)	// Milliseconds sleep

#include <dirent.h>		// opendir closedir readdir scandir
typedef struct  dirent     searchrec;

#define   isDIR(a)    (S_IFDIR==(060000&(a))) // Unix test for directory

				// Adapts attributes  DOS <-- Unix
#define DOSattr(a,n)  ((isDIR(a)?16:0) + ((0200&(a))?0:1) + (('.'==*n)?2:0) \
				+ ((0100&(a))?0:32) /* +((04000&(a))?4:0) */ )

				// Adapts attributes  Unix <-- DOS
#define  UXattr(a)  (((16&(a))?S_IFDIR:0120000) + ((1&(a))?0:0200) + 0444 \
		   + ((32&(a))?0:(uid?0111:0100)) /* +((4&(a))?04000:0) */ )

#else  // not LINUX

#include  <conio.h>		// kbhit
#include     <io.h>		// _findfirst etc
#include <direct.h>		// mkdir rmdir chdir
#include  <sys/utime.h>		// utime
typedef struct _finddata_t searchrec;
#define  ioperm(a,b,c)	1	// Dummy
#define DOSattr(a,n)    (a)	// Dummy
#define  UXattr(a)      (a)	// Dummy
#define __LITTLE_ENDIAN 1234	// e.g. Intel_386
// #define __BIG_ENDIAN 4321	// e.g. Sun_Sparc, Motorola
// #define __PDP_ENDIAN 3412	// (Wow, which PDP model?)
#define __BYTE_ORDER  __LITTLE_ENDIAN

#endif


// Windows file attributes (see <io.h>):
//			[Unix counterpart chosen:]
// _A_RDONLY	0x01	[inverse of "write permission"]
// _A_HIDDEN	0x02	[.fn	    -- too hard to do properly]
// _A_SYSTEM	0x04	[sticky bit -- no, too inconvenient]
// _A_SUBDIR	0x10	[directory]
// _A_ARCH	0x20	[inverse of "executable/walkable"]

// Unix access rights (see <sys/stat.h> , 16bit octals!):
//  S_IFDIR  0040000  directory
//  S_IFCHR  0020000  character oriented file
//	     0060000      Block oriented file
//  S_IFREG  0100000  ordinary file
//	     0004000  sticky bit for owner
//	     0002000  sticky bit for group
//	     0001000  sticky bit for world
//  S_IREAD  0000400     read permission owner
//  S_IWRITE 0000200    write permission owner
//  S_IEXEC  0000100  execute permission owner
//	     0000070   group  permissions
//	     0000007   world  permissions


// _______ more ENDIAN stuff ___________________________________________


  // 'Order4' maps internal byte order (of 4 bytes) to little-endian order:
#if __BYTE_ORDER == __LITTLE_ENDIAN
#define   Order4(i)   (i)	// LittleEndian 0 1 2 3  4 5 6 7  8 9 10 11 ...
#else

#if __BYTE_ORDER == __PDP_ENDIAN
#define   Order4(i)   ((i)^2)	// PdpEndian    2 3 0 1  6 7 4 5  10 11 8 9 ...
#else		//  __BIG_ENDIAN
#define   Order4(i)   ((i)^3)	// BigEndian    3 2 1 0  7 6 5 4  11 10 9 8 ...
#endif

#endif

// _____________________________________________________________________


typedef unsigned short word;
typedef unsigned char  byte;
typedef		 int   boolean;

// #define  BLOCKSZ	4096		 // 4K (bytes)
#define     BLOCKLAST	4100		 // 4 extra bytes for CRC

// byte blockheader[]= {0x18,0x24,0x42,0x81}; // (Testing all bits)
#define BLOCKHEADK   ((long)0x81422418)	 // The header, little_endian

typedef struct {
  long blockheader			;// <-- Will be set by called function
  long blocksize			;
  long blocknr				;
  char blockdata[BLOCKLAST]		;// Incl. CRC at the very end
} databuftype;

#define parerr_ok	0
#define parerr_file	1
#define parerr_dir	2
#define parerr_aborted	3
#define parerr_timeout	4
#define parerr_fatal	5
#define parerr_inuse	6
// intermixed with:  success:1   failure: -1   (cf ReceiveCommandString)

// _____________________________________________________________________

static char    portnr[] = {0,0}		;// Command parameter, initially ""
static boolean blocktransfer		;// Currently in a block transfer?
static boolean abortClicked		;// Checks for any key
static int     portaddr, portaddr1	;// Address of parallel port
	    // portaddr+0 is for Write, +1 is for Read, +2 is ControlReg
	    // (ControlReg gets used only once: for disabling interrupts)

static char    scratch[750 /*256*/]	;// For building textual messages
static databuftype  indbuf;
static databuftype outdbuf = { BLOCKHEADK };

static int     doserror  = 0		;// See <errno.h> on Windows

   // Linux:
// struct dirent {			 // see: dirent.h + bits/dirent.h
//   __ino_t		d_ino;		 // unsigned long
//   __off_t		d_off;		 //   signed long
//   unsigned short int d_reclen;
//   unsigned char      d_type;
//	      char      d_name[256];
// }

   // Windows:
//struct _finddata_t {			 // see:  io.h
//   unsigned attrib     ;
//   time_t   time_create;		 // -1 on FAT file systems
//   time_t   time_access;		 // -1 on FAT file systems
//   time_t   time_write ;		 // (long)
//  _fsize_t  size	 ;		 // (unsigned long)
//   char     name[260]  ;
//}

// In Unix, and in all versions of Microsoft C/C++ except one,
// the time() functions count the seconds from midnight before 1-Jan-1970.
// The time counter will overflow (32 bits!) in 2038.
// Only Microsoft C/C++ Version 7 counts from midnight before 31-Dec-1899.
// The "file date" is date and time of last change (time_write).


// _____________________________________________________________________

#if  defined(LINUX)
#define anyfile  "*"
#else
#define anyfile  "*.*"

// Definition of procedures which are obviously unknown in Windows:

// void outb(unsigned char,unsigned short);	// in sys/io.h
// unsigned char inb(unsigned short);		// in sys/io.h

// For UID and GID see:   <sys/stat.h>     _stat()   _fstat()

#define  getuid()	1		// Dummy
#define geteuid()	1		// Dummy
#define seteuid(a)	(void)a		// Dummy
// The concept of *effective* uid doesn't exist in Windows

#endif

// _______ Adapting the character sets _________________________________

// Code conversion (e.g. ANSI8 / UTF8) is applied in two places:
//    (A) in the protocol, specifically file and directory names,
//    (B) inside text files (which may hence change length)

// Feature:
// On (B) only, and if the DOS character set has been recognized, we convert
// 'high' MS-DOS characters to ANSI8 (notably national characters,
// but not box graphics), never to convert back to MS-DOS though.
// (DOS treatment can be forced by placing a non-ANSI char near start of file.)

// Sequence:  [DOS -->]  Windows_ANSI8 --> Linux_ANSI8  [--> Linux charset]
//			 Windows_ANSI8 <-- Linux_ANSI8  [<-- Linux charset]


static int DOSmode=0;		// Must get set to zero at each APUT


   static char
Dos2Ansi(int c)		// Converts %c:  DOS-or-ANSI8  --> Windows_ANSI8
{  //'Dos2Ansi' is of use only when moving files from Windows_PC to Linux_PC.

   static char dosTable[] = {	// High_DOS to ANSI8 translation table
     199,252,233,226,228,224,229,231,234,235,232,239,238,236,196,197,
     201,230,198,244,246,242,251,249,255,214,220,162,163,165,182,'f',
     225,237,243,250,241,209,170,186,191,173,172,189,188,161,171,187,

				// (This segment is out==in : box graphics)
     176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,

     164,223,245,'n',253,254,181,248,168,226,227,240,236,216,128,'^',
     175,177,169,174,166,167,247,'~',176,185,184,190,179,178,183,160  };

   if	     (c & 0x80)
   { byte u = c & 0x7F;

     if (!DOSmode)
     switch (u) {	// Filtering out normal ANSI characters
       default:  if (u < 0x20) { DOSmode=1; break; }	// DOS identification
       case  0:  case 0x11:  case 0x12:  goto X;	// Euro Lquote Rquote
     }
     return dosTable[u];
   }
X: return (char)c;
}


// _______ UTF8 stuff, and special case Euro ____________________________

   // The transfer protocol uses the 8 bit ISO_8859-1 (=ANSI) character set;
   // it is used by  Windows-95/98/ME.
   // (Windows NT/2000/XP uses Unicode internally, though the conventional
   //  file interface uses 8 bit ANSI8 characters, for compatibility.
   //  There are corresponding 'wide' 16 bit functions.)
   // UTF8 is a Unicode (UCS2) encoding.  The first 256 Unicode characters
   // coincide with ISO_8859-1.
   // TC_SERV offers the USING_UTF8 flag for the case of your Linux using UTF8.

   // Since the Euro symbol was created only very late (~1996), system
   // designers came up with very different ideas on how to accomodate it.

   // The original 8-bit ANSI (= ISO8859-1 = Latin-1) made no provision for
   // the EURO symbol.  That seems to be the state of affairs at Red Hat
   // Linux, at least up to version 9 (Red Hat 9 is not UTF8 based).
   // Windows uses ISO8859-1 but with 0x80 added as its EURO symbol
   // (except in Windows Cyrillic, where it is 0x88).
   // 0x80 is used throughout the TC_SERV program.

   // ISO_8859-15 (= Latin-9, still 8-bit, this is the European extension)
   // has a proper Euro symbol: 0xA4 ;
   // this is used in SuSE Linux up to version 9.0 .
   // With version 9.1 , SuSE switched to UTF8.

   // As of version 2.1 , Unicode (UCS2) has a Euro: 0x20AC (UTF8: E2 82 AC)

#define EURO8  '\200'	 // 0x80 Windows ANSI8 Euro symbol
#define EURA4  '\244'	 // 0xA4 ISO_8859-15   Euro symbol, before SuSE9.1 only
#define EUR1   '\342'	 // 0xE2 1st UTF8
#define EUR2   '\202'	 // 0x82 2nd UTF8
#define EUR3   '\254'	 // 0xAC 3rd UTF8
   //      The '\...' notation is sign extended iff your 'char' is signed
   // (which it mostly is).  This notation works fine (e.g. equality tests)
   // in a consistent 'char' and 'int' context, but it becomes error prone
   // once you mess around with the high bits, e.g. after 'c & 0xFF' .


#if !defined(USING_UTF8) // (it is then ISO_8859_15 or ANSI)
#if !defined(USING_ANSI) // (All characters are unchanged with ANSI systems)


   static char
ISO8859_15(char c)	 // Swaps    %c:   ISO_8859-15 <--> Windows_ANSI8
   { switch (c) { case EURO8: case EURA4: c=(EURO8+EURA4)-c; }; return c; }
   // In the ISO_8859-15 --> Windows direction you may wish to convert
   // (from Czech inverse-hat = "Hatchek" = "Caron" characters):
   //	A6->S  A8->s  B4->Z  B8->z ,  furthermore  BC,BD,BE -> space


   static void
FromUTF8(char *s)	// Swaps    %s:   ISO_8859-15 <--> Windows_ANSI8
   { do { *s= ISO8859_15(*s); } while (*(s++)); } // String length is unchanged


#else  // USING_ANSI
#define   ISO8859_15(c)	(c)	// Dummy for ANSI
#define   FromUTF8(s)		// Dummy for ANSI
#endif // USING_ANSI

#define   ToUTF8 FromUTF8	// (used for file names only)

#else  // USING_UTF8


#define   ISO8859_15(c)  (c)	// Dummy for non ISO8859-15 Linuxes
       // Euro conversion is now done deep in the UTF8 converters


   static void
ToUTF8(char *s)		// Converts %s in-place: UTF8 <--  Windows_ANSI8
{			// String may get longer (up to double length)
   char  c;		// ( make sure 's' is big enough ! ).
   char *from = s;
   char *to   = s+1;	// Will be pointing beyond final '\0'

   while ((c= *(from++)))  to += (c & 0x80) ?(EURO8==c ?3 :2) :1;
			// to = &s[strlen_UTF8(s)]

   while (to != from) {
     if  (0x80 & (c = *--from)) {
      if (EURO8== c)  { *--to= EUR3; *--to= EUR2; c=EUR1; } else
      { *--to =   c & 0xBF  ;
	    c =  (c & 0x40) ? '\303' : '\302';
      }
     }
     *--to    =   c;
   }
}


   static void
FromUTF8(char *s)	// Converts %s in-place: UTF8  --> Windows_ANSI8
{			// String may get shorter
   // Only character codes 0...255 can be mapped to ANSI8 (= 8 bits).
   // Everything else gets mapped to '~'.

   int one, two;
   byte *from  = (byte*)s;
   byte *to    =  from;

   do { *(to++) =     0 ==(0x0080 &(one = from[0]))   ?  one	     :
		 0xC280 ==(0xFEC0 &(two = from[1]     | (one << 8)))
					?(from++, two | (one << 6) ) :
		 0xC080 ==(0xE0C0 & two)?(from++,  '~')		     :
		 0xE282 == two   && EUR3==(char)from[2] ?(from +=2,  EURO8):
	(from += 0xE080 ==(0xF0C0 & two) && 0x80==(0xC0 & from[2]) ?2:0, '~');
      } while  (*(from++));
}

#endif // USING_UTF8


// ____ UTF8 --> ANSI8 conversion while reading a file _________________


static char  ch = '\0'	;// The character to be stored
static int   ci =   0	;// The character read  (EOF recognition:  EOF == ci )
			 // Must get cleared before each fopen and each fseek
static FILE *fi = NULL	;// Note:  'fo' is shared with 'fi'
#define  GETci()  (ci=getc(fi)) // Reads one character from 'fi' into 'ci'
	// Note: getc delivers ALL characters in & 0xFF style;  EOF is -1


#if !defined(USING_UTF8)
#define MapUTF8()	 // Dummy
#else

// Illegal bytes get mapped to one '~'.
// Legal 2-byte or 3-byte UTF8s get mapped to one character, possibly '~'.
// As a side effect, 1 or 2 additional GETci may get called.
// Consequence:  EOF != ci  must get checked before subsequent GETci


   static void
MapUTF8(void)		 // Maps UTF8 --> ANSI8  (for textMode only)
{	// Typically following after:    GETci();  ch=(char)ci;

   if (0==(ch &0x80))  return;

   // Now, proper UTF8 characters are like this:
   //    110x xxxx  followed by one 10xx xxxx	// 2 byte UTF8 --> 11 bit info
   // or 1110 xxxx  followed by two 10xx xxxx	// 3 byte UTF8 --> 16 bit info

   if (0==(ch &0x40) || ci==EOF || 0x80 !=(0xC0 & GETci())) {ch='~'; return;}
   // Improper stuff || was EOF ||    e.g. ci == EOF

   // Now:	(ch &0xC0)==0xC0 , and (ci &0xC0)==0x80 but ci is not EOF

   if  (0x20 == (0x30 & ch))			// 3 byte UTF8 character
   {
     char h =	(char)ci;
     ch = EUR3==(char)GETci()  &&  EUR2==h  &&  EUR1==ch  ?EURO8 :'~';
     return;
   }

   ch = 2 != (0x3E & ch) ?'~' :(ci |(ch << 6));	// 2 byte UTF8 character
}

#endif // USING_UTF8


// ________ Console reports ____________________________________________


static int atmLen = 0	;// Length of previous string
static int atmNL  = 1	;// Flag: previous char was NL
   // The cursor waiting at the end of the line looked too strange;
   // hence 'atmNL' was introduced.



   static void
AddToMemo(char *s)	 // Prints a message on the console
{
   if ('\n' == *s) { atmLen= -1;  if (atmNL) { s++; atmLen++; } }
   atmNL   = 0;
   atmLen += strlen(s);  // 'strlen' goes wrong for UTF8 but works for ANSI8.

   strcpy(scratch,s);
   ToUTF8(scratch);
   printf(scratch);
   fflush(stdout);
}
   // 'AddToMemo' is a bit odd, in that its argument string 's' must
   // be in ANSI8 (not in UTF8).  -  All comments in TC_SERV are
   // in English, i.e. they don't contain characters with bit 0x80 set.
   // The protocol echo (at the beginning of DoServer) is in ANSI8.


   static void
AddToMemoOkay(void)	 // Writes 'OK' on the console
{
   int  i = 48 - atmLen%80;
   atmLen = 0;
   atmNL  = 1;
   do { putchar(' '); } while (0 < --i);
   printf("OK\n");
}


// ______ Date format conversion _______________________________________

// struct tm
//   int tm_sec		;// [0, 59] seconds after the minute
//   int tm_min		;// [0, 59] minutes after the hour
//   int tm_hour	;// [0, 23] hours   since midnight
//   int tm_mday	;// [1, 31] day of the month
//   int tm_mon		;// [0, 11] months  SINCE January	(!)
//   int tm_year	;//	    years   since 1900
//   int tm_wday	;// [0,  6] days    since Sunday
///  int tm_yday	;// [0,366] days    since January 1
///  int tm_isdst	;//	    daylight savings time flag


   static time_t
UXtime(long fT)			// Adapts dates&times  Unix <-- FAT32
{
   // Note: FAT32 date&time format has many "illegal" values (e.g. fT=0 means
   // 00.00.1980 at 0:00 hrs ! -- Unix expresses legal dates&times only).
   // These will be converted into Unix surprise values, or 31-Dec-1971 0:00.
   // This should cause no massive problem - dates&times rarely matter!
   // TC deals mostly with legal dates&times.
   fat_time.L      = fT;
   ux_time.tm_sec  =(fat_time.f.secs*59+15)/31 ;
   ux_time.tm_min  = fat_time.f.mins;
   ux_time.tm_hour = fat_time.f.hrs ;
   ux_time.tm_mday = fat_time.f.day ;
   ux_time.tm_mon  = fat_time.f.month-1;
   ux_time.tm_year = fat_time.f.year +80;
   { time_t x=mktime(&ux_time);	// 'mktime' converts   time_t <-- time_s
				//  with (time_t)-1  as error indicator
  // Converting future to past: (though, how many of your files are pre-1980 ?)
     if ((time_a.time+100000)<x) { ux_time.tm_year -=128; x=mktime(&ux_time); }
     return x;
}  }


   static long
FATtime(time_t uT)		 // Adapts dates&times  FAT32 <-- Unix
{				 // Wraps around if time  T < 1-Jan-1980
   ux_time = *localtime(&uT)	;// 'localtime' converts  time_s <-- time_t
   fat_time.f.secs  =(ux_time.tm_sec*31+27)/59;  // 2 secs is max fat resolution
   fat_time.f.mins  = ux_time.tm_min ;
   fat_time.f.hrs   = ux_time.tm_hour;
   fat_time.f.day   = ux_time.tm_mday;
   fat_time.f.month = ux_time.tm_mon +1;
   fat_time.f.year  = ux_time.tm_year-80;
   return fat_time.L;
}
// With casts (instead of the union) we got the warning (gcc -Wall)
// "dereferencing type-punned pointer will break strict-aliasing rules"


// _____________________________________________________________________

// Replacement for 'fwrite'
//   Its definition is -essentially- like the standard


   static long			// Augmented replacement for 'fwrite'
my_fwrite(char* buf, long busize, int textMode, FILE *f)
{  // Notice:  'sizo' got replaced by 'textMode'
   long i=0;

   if    (!textMode)
     for (;  i != busize;  i++)  putc(buf[i],f);	// Writing binary data
   else  {						// Writing   text data
     char ch;
     for (;  i != busize;  i++)
      if ('\r' != (ch = buf[i]) )
       {
	  ch = ISO8859_15(Dos2Ansi(ch));

#if defined(USING_UTF8)
	  if (0x80 & ch) {
	    if (EURO8==ch) { putc(EUR1,f);  putc(EUR2,f);  ch=EUR3; } else
			   { putc((0x40& ch) ?'\303':'\302', f);  ch &= 0xBF; }
	  }
#endif // USING_UTF8

	  putc(ch, f);
       }
   }

   return busize;
}


// _____________________________________________________________________


#if defined(LINUX)

static  int       keyflag = false;
#define kbhit()	  keyflag	// Tests if abort has been requested


   static void			// type == SIGnal
SignalHandler(int type)		// Processes a signal, ensures clean exits
{
   keyflag=true;

   switch (type) {
     case SIGSEGV: printf("\n\n*** Serious fault in TC_SERV !\n"); exit(1);
		   // Segmentation fault!

     case SIGINT:  // Ctrl-C
		   AddToMemo("\n*** You pressed Ctrl/C");
		   return;
 //  case SIGHUP :
 //  case SIGTERM:
 //  default     : break;
   }
   printf("\n*** Signal %u received\n",type); fflush(stdout);
   exit(0);
}

#endif   // LINUX


   static byte
ReadPort(int pp)	// Reads  a byte from an IO port
{
   byte result;
   word p = (word)pp;
#if defined(LINUX)
   result = inb(p);
#else
   __asm
  {
    mov dx,p
    in  al,dx
    mov result,al
  }
#endif
  return result;
}

   static void
WritePort(int pp, int bb) // Writes a byte   to an IO port
{
   word p = (word)pp;
   byte b = (byte)bb;

#if defined(LINUX)
   outb(b,p);
#else
  __asm
  {
    mov dx,p
    mov al,b
    out dx,al
  }
#endif
}


   static int
GetPortAddress(int lptportnr)	// Delivers address of parallel interface
{			// lptportnr:  1 for LPT1,  2 for LPT2,  3 for LPT3
   int result;

   switch (lptportnr) {
 //  DOS only: result= *(short*)(0x406 + 2*lptportnr);  break;
     case  1:  result= 0x3BC;  break;	// (trad. Hercules printer port)
     default:  portnr[0]='2';
     case  2:  result= 0x378;  break;	// (quasi standard)
     case  3:  result= 0x278;  break;
   }

   return ioperm(result,3,6) ?0 :result; // 3 bytes; permission 6: read+write
   // 'ioperm' is mostly for the i386 architecture.
   // On other architectures it does not exist or will always return an error.
   // Note, non-existing printer ports do NOT cause an ioperm error.
   // Hence we have to identify the port by listening for input.
}

// ________________________ Low level functions _________________________

#define  WriteReg0(a)  WritePort(portaddr,a) // Writes write port 0
#define   ReadReg1()	ReadPort(portaddr1)  // Reads  read  port 1
#define PortStatus()	(0x80 & ReadReg1())  // Gets   read  port 1 test bit
  // In   'Upload' direction C<--S, bit 0x10 gets wiggled up and down (Flag16)
  // In 'Download' direction C-->S, bit 0x80 gets wiggled up and down
#define SendAck()	WriteReg0(0)	     // Sends "Ack"
#define SetPortReady()	SendAck()	// Sends/sets "Ready" (same as "Ack")

#define NACK		15	// Negative Acknowledge, please resend
#define ABORT		14	// Abort entire transmission
#define AFILE		13	// Abort transfer of this file, no retry

   // Protocol codes for  WritePort(portaddr, code) :
   //    0   Ready or Ack
   //    1   Error (usually followed by Abort)
   //   13   Abort transfer of this file, no retry
   //   14   Abort entire transmission
   //   15   NAck, please resend
   //   16   Flag16 or:  Low nibble (to be followed by High nibble)
   //	     All codes <16 occur also in High nibble transmission


   static boolean		// Waits for status bit flip or 2 sec timeout
WaitForBitFlip(boolean bithigh, int starttime)
{				// Result:  true:success   false:failure
// int starttime = 21;		// Testing for max 2 secs altogether
   while   (0< --starttime)
   {
     int tries = 10000;		// Max 10000 bit tests each time
     while (0< --tries)  if (bithigh == (0 != PortStatus()))  return true;
    _sleep(100L);		// 0.1 secs pause
   }		// (Without sleeps the server would block the computer.)
   return false;
}		// Description:  WaitForBitFlip keeps trying for max 2 secs



   static void
SendDisableInterrupts(void)	// Disables interrupts
   { WritePort(portaddr+2, 4); }// Interrupt control port
		// The entire program will never re-enable interrupts.


   static void
SignalConnectionError(void)	// Transmits a Connection Error signal
{
   WriteReg0(ABORT   );  _sleep(50L);	// Abort      + 1/20 sec
   WriteReg0(ABORT+16);  _sleep(50L);	// Abort+16   + 1/20 sec
   WriteReg0(ABORT   );  _sleep(50L);	// Abort      + 1/20 sec
   WriteReg0(ABORT+16);  _sleep(50L);	// Abort+16   + 1/20 sec
}

   static void
SignalAbortTransfer(void)	// Transmits a Transfer Abort request
{
   WriteReg0(ABORT+16);  _sleep(100L);	// Abort+16   + 0.1 sec
   WriteReg0(ABORT   );  _sleep(100L);	// Abort      + 0.1 sec
}


   static long
StrToInt(char *s)	// Reads "%10lu" from a string;  -1 indicates error
{
   char *end = s;
   long    i = strtoul(s, &end, 10);
   return  (s + strlen(s)) != end    ? -1   :i ;

// char  *p;
// for   (p=s; *p ; p++)  if (*p < '0'  ||  '9' < *p)  return -1;
// sscanf(s, "%lu", &i);  // Not quite as safe -- no length test
// return i;
}

// ________________________ Send/receive data _________________________

// 'UploadBlock()' sends TO Total Commander client
// (btw: failure report is "DOWNLOAD FAILED" (!) )
// All 'data', except CRC, must be supplied by caller.

   static int
UploadBlock(databuftype *data)	// Sends one 'data' block   to parport
{				// Result:  one of the parerr_  codes
   int     i;
   int     size    = data->blocksize+16; // Size incl. header and with CRC
   int     sizem4  = size-4;		 // Size incl. header without  CRC
   long    crc32       ;
   word    crc32_1     ;
   word    crc32_2     ;
   byte   *pdata   = (byte*)data;	 // [size]
   byte   *pcrc    = (byte*)&crc32;	 // [4]
   byte    thisChar;
   boolean ackreceived = true;
   int     servererror = parerr_ok;

   SendDisableInterrupts();
   data->blockheader = BLOCKHEADK;

   /* Block layout: 4 byte header,  4 byte block size,  4 byte blocknr,
				    block data,		4 byte crc */
again:
   crc32_1 =  1;
   crc32_2 =  0;
   crc32   = -1;
   i       =  0;

  while     (i<size && ackreceived) {

    if      (i<12)		// (#1) Sending the header (first 12 bytes)
      thisChar = pdata[Order4(i)];

    else if (i < sizem4) {	// (#2) Sending the blockdata proper
      thisChar = pdata[i];
      crc32_1 += thisChar;	// CRC only for the blockdata proper.
      crc32_2 += crc32_1;	// This CRC is unconventional.
    }				// That's okay, as long as it delivers safety
				// and if both sides use the same algorithm.
  /* Comment from Christian Ghisler:
     This is the "Adler32" checksum (named after Marc Adler, one of
     the Infozip developers) and it is used in zlib .  Especially on
     slow computers, this checksum is markedly faster than CRC32,
     without being less effective.
   */

    else {			// (#3) Sending the CRC    ( last  4 bytes)
      if    (sizem4 == i)   crc32 = (crc32_2 << 16) + crc32_1;
      thisChar = pcrc[Order4(i-sizem4)];
    }

    WriteReg0(0x10 | (thisChar & 0xF));		// Low nibble
    ackreceived   = /* 0 == PortStatus() || 0 == PortStatus() || */
		    WaitForBitFlip(false, 20);
				// Why this frantic PortStatus testing ???

    if (ackreceived) {
      WriteReg0((thisChar >> 4) & 0xF);		// High nibble
      ackreceived = /* 0 != PortStatus() || 0 != PortStatus() || */
		    WaitForBitFlip(true, 20);
    }

    if   (ackreceived) {

      switch ((ReadReg1() >>3) & 0xF) {	// Only data part!
	case NACK:	// 15: Please resend the entire block
	  goto again;

	case ABORT:	// 14: Entire transmission aborted
	  ackreceived = false;
	  servererror = parerr_fatal;
	  break;

	case AFILE:	// 13: Transmission aborted (for one file)
	  ackreceived = false;
	  servererror = parerr_file;
	  break;
      }
      i++;
    }
  }

  if (! ackreceived) {
    SetPortReady();
   _sleep(50L);		// 1/20 sec
  }

  return   servererror != parerr_ok ? servererror
	: (ackreceived ?  parerr_ok : parerr_timeout);
}  // UploadBlock


// 'DownloadBlock()' receives FROM Total Commander client
//(btw: failure report is "UPLOAD FAILED" (!) )
// Only 'data->blocknr' must be supplied (and it will remain unchanged)


   static int
DownloadBlock(databuftype *data)  // Reads one 'data' block from parport
{				  // Result:  one of the parerr_  codes
   boolean want_more;
   int     result      = parerr_ok;
   int     retransmits =  0;
   byte   *rawdata     = (byte*)&data->blockdata[-12];	// [BLOCKLAST+12]
   long    blocksz     =  0;
   long    sentblocknr =  0;
   byte    thisChar	   ;	// Collecting a byte out of two nibbles
   long    thisLong    =  0;	// Collecting a long out of four bytes
   int     dataidx     =  0;
   word    crc32_1     =  1;
   word    crc32_2     =  0;

   SendDisableInterrupts();

   do {

     want_more = /* 0==PortStatus() || 0==PortStatus() || */
		 WaitForBitFlip(false,20);
     if (! want_more) { result = parerr_fatal;  continue; }

     thisChar  = (byte) ((ReadReg1() >> 3) & 0x0F);	// Low nibble

     WriteReg0(16);					// Flag bit 16
     want_more =    /* PortStatus() ||    PortStatus() || */
		    WaitForBitFlip(true,20);
     if (! want_more)  { result = parerr_fatal;  continue; }

     thisChar += (byte) ((ReadReg1() << 1) & 0xF0);	// High nibble

     if (11 < dataidx)
     {
	if   ((blocksz+11) < dataidx)	  // (#3) Collecting transmitted CRC
	     thisLong = (thisChar<<24) + (((unsigned long)thisLong)>>8);
	else
	{    // (#2) Collecting data proper, also recalculating CRC on the fly
	     rawdata[dataidx] = thisChar; // not: [...-12]
	     crc32_1 += thisChar;
	     crc32_2 += crc32_1;
	}
     }
     else {  // (#1) Collecting header (12 bytes), though synchronizing first
	thisLong = (thisChar<<24) + (((unsigned long)thisLong)>>8);

	switch (dataidx) {
	  case 11:     sentblocknr =  thisLong;  goto clear;
	  case  7:	   blocksz =  thisLong;  goto clear;

	  case  3:  if (BLOCKHEADK == thisLong)	// Perfect match
clear:			{ thisLong=0;  break; }

			// Typical error: blocks out of sync.
			// Checking if out of step by 1 byte.
		    if (0==((thisLong-(BLOCKHEADK<<8)) & 0xFFFFFF00)) {
			dataidx=2;
			AddToMemo("\n... Synchronizing --");
			break;
		    }

		    if (0==((thisLong-(BLOCKHEADK>>8)) & 0xFFFFFF)) {
			// (We never saw BLOCKHEADK's initial byte)
			thisLong &= 0xFF000000;
			dataidx=4;
			AddToMemo("\n... Synchronizing ++");
			break;
		    }

		    result    = parerr_fatal;
		    want_more = false;
		    break;
	}
     }

       ++dataidx;
     if (dataidx != (blocksz+16))   { SendAck();  continue; }

     // Checking everything, once the data proper have been read:
	    // Couldn't we have checked 'sentblocknr' BEFORE reading data?
     if		(sentblocknr== -1) {	 // Request of an Abort
	    WriteReg0(AFILE); // OK, will abort transfer of this file, no retry
	    result = parerr_aborted;
     } else if  (thisLong != ((crc32_2 << 16) + crc32_1)) {
	    retransmits++;		 // Wrong crc --> try again
	    WriteReg0(NACK);		 // 15: Data bad - causes block resend
	    if  (10<retransmits)  result = parerr_fatal;
     } else if  (sentblocknr != data->blocknr) {
					 // Wrong block -> fatal error
	    WriteReg0(data->blocknr== -1 // Should never be encountered:
			?AFILE		 // Aborting transfer of file, no retry
			:ABORT );	 // Entire transmission aborted
	    result = (data->blocknr== -1 ||  abs(sentblocknr-data->blocknr)< 3)
		  ? parerr_aborted	 // Error at block end
		  : parerr_fatal;
     } else {
	    SendAck();
	    // Storing the 'long' items:
	    *(long*)&rawdata[dataidx-4]= thisLong;  // Storing received CRC
	    data->blockheader = BLOCKHEADK;
	    data->blocksize   = blocksz    ;
	 // data->blocknr     = sentblocknr;	    // Not storing this one!
	    retransmits       = 0;
	    want_more	      = false;		    // Success!   All done.
     }
     crc32_1  = 1;
     crc32_2  = 0;
     dataidx  = 0;
     thisLong = 0;
   } while (want_more  &&  result == parerr_ok);

   if (result==parerr_fatal) {
     SignalAbortTransfer();	// Abort transfer
     SetPortReady();		// Clear error condition
   }

   return result;
}  // DownloadBlock


// ________________________ Send/receive commands _________________________


   static boolean
RemoteSendCommand(char *command)  // Sends    a text block   to parport
{		// Result: true:success,  false:failure/try_again_later
   int i;
   if (blocktransfer) return false;

   strcpy(outdbuf.blockdata, command);
   outdbuf.blocksize = strlen(command);
   outdbuf.blocknr   = 0;

   blocktransfer = true;
	       i = UploadBlock(&outdbuf);
   blocktransfer = false;

// printf("\nRemoteSendCommand(\"%s\");\t-->  %d", command, i);

   return parerr_ok==i;
}


   static int
ReceiveCommandString(char *p)   // Reads  one command from parport
{			// Result: 1:success, 6:try_again_later, -1:failure
   int   i;
   int   result = -1;

   if (blocktransfer)  return parerr_inuse;

   blocktransfer  = true;
   indbuf.blocknr = 0;
   if (parerr_ok==DownloadBlock(&indbuf)) {
     for (i=0;  i < indbuf.blocksize ;i++)  *(p++)= indbuf.blockdata[i];
     result = 1;
   }
   *p = '\0';
   blocktransfer  = false;
   return result;
}

// ___________________ Receive file (client or server!) ____________________


   static char*
FormatUnixName(char *filenam)	// Adapts filenames  Unix <-- Dos
{  // Leaves its argument unchanged.  (This is intentional.)
   static char  filename[/*350*/ 700];
   int i;

   switch (*filenam) {		// Makes all filenames absolute
     case '\0':  strcpy (filename, "/"			); break;
     case '/' :  strcpy (filename,	  filenam	); break;
     case '\\':  filenam++;
     default  :  sprintf(filename, "/%s", filenam	); break;
   }
				// Changes each backslash to forward slash
   for (i= (int)strlen(filename) ;   0 < --i  ;)
     if (filename[i]=='\\') filename[i] = '/';

   i = strlen(filename)-1;	// Removes any trailing slash
   if (0<i && '/'==filename[i]) filename[i] = '\0';

//printf("\nFile \"%s\"", filename);

   ToUTF8(filename);
   return filename;
}

   static void
FlopMessage(void)	// Reports "Cannot allocate file"
{  AddToMemo("\nCannot allocate file -- No write permission or device full"); }


// Byte counts on the (Windows) client side:
//     restartAtRemote, filesize, totalreceived, indbuf.blocksize,
//     (result of:) GetRestartAtRemote, (argument of:) REST=
// Byte counts on the (Linux) server side:
//     restartAt


   // Receives a file FROM THE CLIENT;  Result:  parerr_<whatever>
   static int
ReceiveData(char *filename,
	long filesize  ,long filetime	     ,long fileattr,
	long restartAt ,long restartAtRemote ,boolean textMode)
{					// -- has exactly one reference --
#define    fo   fi
// FILE   *fo = NULL;			// There is only 'fi' ...
   static char  ufilenam [512];
// static char  textData [BLOCKLAST+1-4];
   long    totalreceived, blocknr;
   int     err, ioerr, i;
   boolean want_more, compareOK;

   int     result = parerr_ok;

   if (blocktransfer)  return parerr_inuse;

   blocktransfer = true;

   // Name adaptation of "hidden" files (add or take a dot) will cause chaos!
   // To illustrate this:  ServerUploadFile (doesn't get properties)
   // would be unable to decide if it should load file 'fred' or '.fred'

   strcpy(ufilenam, FormatUnixName(filename));

   if (filesize==restartAtRemote) {	// (ptp)
      // Create a new empty file (write file even though empty)
      ci = 0;
      if (0==filesize  &&  NULL != (fo=fopen(ufilenam, "wb")))  goto closefile;
      blocktransfer = false;
      FlopMessage();
      return parerr_file;		// "Cannot allocate file for writing"
   }

   totalreceived = restartAtRemote;
   blocknr       = 1;

   do {
      indbuf.blocknr = blocknr;

      err = DownloadBlock(&indbuf);
      if (parerr_ok != err)
      {
	printf("\n*** %lu Error", totalreceived);
	FlopMessage();
	result    = err;
	want_more = false;
	continue;
      }
      want_more = true;

      if   (1==blocknr  ) {	// Data OK -> Open local file
	ci = 0;
	if (0==restartAt) {	// Create a new file (and write to it)
	  ioerr = NULL == (fo=fopen(ufilenam, (textMode?"w":"wb")));
	} else {		// Now: restartAt > 0
	  ioerr = NULL == (fo=fopen(ufilenam, (textMode?"a+":"a+b")));
				// Open an existing file for reading+appending
	  if (!ioerr) {
	    ci = 0;
	    ioerr = fseek(fo,restartAt,0);	// Repositioning
	    if (!ioerr) {
	      compareOK = true;

	      if (textMode) {
		for (i=0; i<indbuf.blocksize && EOF != GETci(); i++) {
		    ch = indbuf.blockdata[i];
		    if ('\r' != ch) {		// Skipping CRs

		      ch = ISO8859_15(Dos2Ansi(ch));
#if defined(USING_UTF8)
		      if (0x80& ch)
		      {
			if (EURO8==ch)
			{
			  if (EUR1 != (char)ci)  goto P;
			  if (EOF  ==  GETci())  break ;
			  if (EUR2 != (char)ci)  goto P;
			  ch= EUR3;
			} else
			{
			  if (((0x40& ch) ?'\303':'\302') != (char)ci) goto P;
			  ch &= 0xBF;
			}
			if (EOF == GETci())  break;
		      }
#endif // USING_UTF8

		      if (ch != (char)ci) goto P;
		    }
		}
	      } else  // Binary mode:
		for (i=0; i<indbuf.blocksize  &&  EOF != GETci()  ;i++)
		    if ( (ch= indbuf.blockdata[i]) != (char)ci) {
P:			 printf("%u:%02X-%02X   ",i,ch,ci);
			 compareOK=false;  break;
		    }
	      // In real, we would have liked to flag an error here

	      if (! compareOK) {
		AddToMemo("\n*** Error: source and target differ!");
		blocknr = -1;
		abortClicked = true;
	      } else {
		ci = 0;
		ioerr = fseek(fo,restartAt,0);	// Repositioning
	      }
	    } // if (!ioerr)
	  }   // if (!ioerr)
	}     // if (0==restartAt) ...
	if (ioerr) blocknr= -1;	// Send blocknr== -1 as abort signal

      }    /* if(1==blocknr) */	// Will abort in next block

      if (blocknr == -1)  continue;

      blocknr++;
      totalreceived += indbuf.blocksize;
      (void) my_fwrite(indbuf.blockdata,indbuf.blocksize,textMode,fo);
      if (0 == (blocknr%25))  AddToMemo("~");  // One ~ every 100 KB
      result = parerr_ok;

  } while (want_more  &&  totalreceived != filesize);


closefile:

  if (fo)
  {
    (void) fstat(fileno(fo), &fprops);
    if (fileattr != -1)  fprops.st_mode  = (unsigned short) UXattr(fileattr);
	// How does one SET the attributes ???   Just by fclose ?
 // if (filetime != -1)  fprops.st_mtime = UXtime(filetime);
	// At 'fclose(fo)',  st_mtime gets overwritten by current time.

    fclose(fo);  fo=NULL;

    if (filetime != -1)
    {
      static struct utimbuf utimbuffer;

/*    --   debug printout  --
 fat_time.L = filetime;
 sprintf(scratch,"\tTimestamp %lu --> %02u.%02u.%04u %02u:%02u.%02u",
 filetime,
 fat_time.f.day, fat_time.f.month, fat_time.f.year+1980,
 fat_time.f.hrs, fat_time.f.mins, (fat_time.f.secs*59+15)/31 );
 AddToMemo(scratch);
 */

      utimbuffer.actime = utimbuffer.modtime = UXtime(filetime);
      (void) utime(ufilenam, &utimbuffer);
    }
  }

  blocktransfer = false;
  return result;
}  // ReceiveData


// ______________ Helpers for Server functions ________________________	_

 /*
  struct stat {
     off_t st_size
    time_t st_mtime
    ushort st_mode
  }
  */

#if  defined(LINUX)

   void			// Uploads -adapted- file info to client
StoreStatus(char *obuf, char *dir, char *name)
{
   word attrib;

   sprintf(scratch,('/'== dir[strlen(dir)-1] ?"%s%s%s" :"%s%s/%s")
		  ,('/'== dir[0] ?"" :"/"), dir, name);

   stat(scratch, &fprops);
   attrib = fprops.st_mode;

// attrib =   (  isDIR(attrib) ? 0x10 : 0 )  // directory
//	    + ((0200 & attrib) ? 0    : 1 )  // write permission / readonly
//	    + (( '.' == *name) ? 2    : 0 )  // hidden
//	    + ((0111 & attrib) ? 0 : 0x20 ); // exec permission / archive

   if ( strcmp(name, "..")  /* &&  strcmp(name, ".") */  )
   {
     FromUTF8(name);
     sprintf (obuf ,"%s\t%lu\t%ld\t%u\r\n"
		   ,name,  (isDIR(attrib) ?0 :fprops.st_size)
		   ,FATtime(fprops.st_mtime), DOSattr(attrib,name));
   }

 /*
   else sprintf(obuf, "%s\t0\t-2\t%u\r\n", name, DOSattr(attrib,"x"));
  */

}

#endif   // LINUX


// __________________________ Server functions ________________________


   static boolean
ServerListDirectory(char *dirnam) // Uploads one directory list
{				  // Result: true:success  false:failure
   static databuftype databuf;
   static char    inbuf[1024];
   static char *dirname /*[ 256]*/;

   long		totaldatasize, startTimeSeconds, timehelp, blocknr;
   int		L, blocktotal;
   typedef long intptr_t;
   boolean      ok;
   searchrec    ss;
   searchrec   *s = &ss;

#if defined(LINUX)
#define Get1Dir(f,name,s)	(s=readdir(f))  // Gets  1st dir entry
#define GetXDir(f,s)		(s=readdir(f))  // Gets next dir entry
#define FAILURE NULL		// Unix Get?Dir error value
   register DIR *f;		// NOT:    register struct dirent *f;

   strcpy(ss.d_name, anyfile);
   ss.d_type     =   0 ;
   ss.d_ino      =   0 ;
   ss.d_off      =   0 ;
   ss.d_reclen   =   0 ;

#else
#define closedir		_findclose
#define Get1Dir(f,name,s)    (f=_findfirst(name, &ss))  // Gets  1st dir entry
#define GetXDir(f,s)		_findnext(f,&ss)	// Gets next dir entry
#define FAILURE (-1)				// Windows Get?Dir error value
#define d_name	name
   intptr_t  f;

   strcpy(ss.name, anyfile);
   ss.size       =   0 ;
   ss.time_write =   0 ;
   ss.attrib     =   0 ;
#endif    // LINUX

   if (blocktransfer) return false;

   dirname = FormatUnixName(dirnam);
   startTimeSeconds = GetEpochTime();

  {	// Determining 'totaldatasize'

#if defined(LINUX)
    doserror = NULL==(f = opendir(dirname));
#else
    doserror = 0;
#endif    // LINUX

    if (doserror  ||  FAILURE == Get1Dir(f,dirname,s)) {
      doserror = ENOENT;   // errno;		 // ENOENT or EINVAL
      return RemoteSendCommand("DIR ERROR 2");
    }

    totaldatasize = 0;
    ok = true;

    while (doserror==0  &&  ok) {
      if  ( strcmp(s->d_name, "..")  /* && strcmp(s->d_name, ".") */ )
      {
#if  defined(LINUX)
	StoreStatus(inbuf, dirname, s->d_name);
#else
	sprintf(inbuf, "%s\t%lu\t%ld\t%u\r\n",
			ss.name, ss.size, ss.time_write, ss.attrib);
#endif    // LINUX
	totaldatasize += (int)strlen(inbuf);
  //    printf("%s ", s->d_name);
      }
      if (startTimeSeconds < (timehelp = GetEpochTime())) {
	  startTimeSeconds =  timehelp;
	  ok = RemoteSendCommand("WAIT");
      }

      doserror = FAILURE==GetXDir(f,s) ?ENOENT /*errno*/ :0;
    } // while doserror==0
    (void) closedir(f);
  }

  sprintf(scratch, "SIZE %lu", totaldatasize);
  ok = RemoteSendCommand(scratch);

  if (ok  &&  0<totaldatasize) {	// Only send data if file != 0
    {
      strcpy(ss.d_name, anyfile);

#if defined(LINUX)
      doserror = NULL==(f = opendir(dirname));
#else
      doserror = 0;
#endif    // LINUX

      if (0==doserror  &&  FAILURE == Get1Dir(f,dirname,s)) {
	 doserror = ENOENT/*errno*/ ;	// 0 or ENOENT or EINVAL
      }

      blocknr       = 1;
      ok	    = true;
      abortClicked  = false;
      blocktransfer = true;
      blocktotal    = 0;

      while (doserror==0) {
	if  ( strcmp(s->d_name, "..")  /* && strcmp(s->d_name, ".") */ )
	{
#if  defined(LINUX)
	  StoreStatus(inbuf, dirname, s->d_name);
#else
	  sprintf(inbuf, "%s\t%lu\t%ld\t%u\r\n",
			ss.name, ss.size, ss.time_write, ss.attrib);
#endif    // LINUX
	  L = (int)strlen(inbuf);
	  if (ok  &&  0<L) {
	    memcpy((char*)&databuf.blockdata[blocktotal], inbuf, L);
	    blocktotal += L;
	    if ((BLOCKLAST-204) < blocktotal) {  // Flush!
	      abortClicked = kbhit();
	      if (abortClicked) {
		databuf.blocksize =  0;
		databuf.blocknr   = -1;
		(void) UploadBlock(&databuf);	// Signal abort
			       ok = false;
	      } else {
		databuf.blocknr   = blocknr;
		databuf.blocksize = blocktotal;
			       ok = parerr_ok==UploadBlock(&databuf);
		blocknr++;
		AddToMemo("~");
	      }
	      blocktotal = 0;
	    }
	  }
	}

	doserror = FAILURE==GetXDir(f,s) ?ENOENT/*errno*/ :0; // 0 or ENOENT
      } // while doserror==0
      (void) closedir(f);
    } // ! getdrivelist

    if (0<blocktotal  &&  ok) {
      blocktransfer     = true;
      databuf.blocknr   = blocknr;
      databuf.blocksize = blocktotal;
		     ok = parerr_ok==UploadBlock(&databuf);
    }
    blocktransfer = false;
  }

  return ok;
}  // ServerListDirectory



   // Transmits a file TO THE CLIENT; Result: true:success  false:failure
   static boolean
ServerUploadFile(char *filename, long restartAt, int textMode)
{
   long  blocknr, theSize=0, theSizeDisk,
	 restartAtDisk, startTimeSeconds, timehelp;
   int   blocki;
// long  fmtime = 0;
// word  fattr  = 0;
   boolean  ok, nextisCR, onlyleadingLF;

   if (blocktransfer) return false;

   ci = 0;
   if (NULL == (fi=fopen(FormatUnixName(filename), (textMode?"r":"rb")))) {
     doserror = ENOENT /*errno*/ ;	// "Cannot open file for reading"
     return RemoteSendCommand("FILE ERROR 2");
   }

   ok = true;
   if (restartAt) {
     ci = 0;
     fseek(fi,restartAt,0);	// Repositioning
     sprintf(scratch, "REST %lu", restartAt);
     ok = RemoteSendCommand(scratch);
   }
   if (ok) {
     long  ts;
     (void) fstat(fileno(fi), &fprops);
 //  fmtime = fprops.st_mtime;
 //  fattr  = fprops.st_mode;
     theSize= fprops.st_size;
     ts     = FATtime(fprops.st_mtime);

/*    --   debug printout  --
 sprintf(scratch,"\tTimestamp %02u.%02u.%04u %02u:%02u.%02u --> %lu",
 fat_time.f.day, fat_time.f.month, fat_time.f.year+1980,
 fat_time.f.hrs, fat_time.f.mins, (fat_time.f.secs*59+15)/31, ts);
 AddToMemo(scratch);
 */

     sprintf(scratch, "TIME %lu", ts);
     ok = RemoteSendCommand(scratch);
   }
   if (ok) {
     sprintf(scratch, "ATTR %u", DOSattr(fprops.st_mode ,filename));
     ok = RemoteSendCommand(scratch);
   }

  onlyleadingLF = false;

  if (textMode)     // Determining the length of a text file:
  {
    theSize       = 0;		// Size on the (Windows) client
    theSizeDisk   = 0;		// Size on the (Linux) server
    restartAtDisk = 0;
    ci = 0;
    (void) fseek(fi,0L,0);			// Repositioning
    startTimeSeconds = GetEpochTime();

    if (restartAt==0) {
      for (; EOF != ci  &&  EOF != GETci()  ;theSize++)
	  { ch=(char)ci;  if ('\n'==ch) theSize++;  else MapUTF8(); }
    } else {		// restartAt position is counted *including* CRLF
      while (EOF != ci  &&  EOF != GETci()) {
	  ch = (char)ci;
	  if   (ch == '\n') theSize++;
	  else MapUTF8();
	  ++theSize;
	  ++theSizeDisk;
	  if (restartAtDisk==0  &&  restartAt <= theSize) {
	    restartAtDisk = theSizeDisk;
	    onlyleadingLF = restartAt < theSize;
	  }
      }
      if (0== (theSizeDisk & 0x7FFF)  &&	// Send WAIT after every 32 KB
	  startTimeSeconds < (timehelp = GetEpochTime())) {
	  startTimeSeconds =  timehelp;
	  ok = RemoteSendCommand("WAIT");
      }
    }
    ci = 0;
    (void)fseek(fi,restartAtDisk,0);		// Repositioning
  }

  if (ok) {
      sprintf(scratch, "SIZE %lu", theSize);
      ok = RemoteSendCommand(scratch);
  }
  if (ok) {
    blocknr       = 1;
    abortClicked  = false;
    blocktransfer = true;
    nextisCR      = false;
    blocki	  = 0;		// = indbuf.blocksize   to be

    // Filling and uploading block after block:

    while  (ok && (EOF != ci  ||  nextisCR)) {

	  // Filling 'indbuf.blockdata' character by character:
	  if   (textMode) {
	    if (nextisCR) {
	      indbuf.blockdata[blocki++] = '\n';
	      nextisCR = false;
	    }
	    else if (EOF != GETci()) {
	      ch = (char)ci;
		  nextisCR = ch=='\n';
	      if (nextisCR)  ch= '\r';
	      if (onlyleadingLF && nextisCR) {
		  blocki--;
		  onlyleadingLF = false;
	      } else {
		MapUTF8();
		indbuf.blockdata[blocki++] = ISO8859_15(ch);
	      }
	    }
	  } else {			// Binary transfer
	      if   (EOF != GETci())
		   indbuf.blockdata[blocki++] = (char)ci;
	   /*
	    static char readblock[32*1024];
	    blocki = min(BLOCKLAST-4, dataRead-i);
	    memcpy(indbuf.blockdata, &readblock[i], blocki);
	    */
	  }

	  // Emptying the buffer (buffer full, or nothing else coming):
	  if (blocki==(BLOCKLAST-4) || (EOF == ci && !nextisCR)) {
	    if ((blocknr%25) ==0)  AddToMemo("~");  // One ~ every 100 KB
	    indbuf.blocknr   = blocknr;
	    indbuf.blocksize = blocki;
			  ok = parerr_ok==UploadBlock(&indbuf);
	    blocki	     = 0;
	    blocknr++;
	    abortClicked = kbhit();
	    if (abortClicked)  ok=false;
	  }
    }     // while (ok && (EOF != ci || nextisCR)) ...

    blocktransfer = false;
  }
  fclose(fi);  fi=NULL;
  return ok;
}  // ServerUploadFile


   static boolean
ServerCreateDirectory(char *dirname) // Creates a directory on the server
{
#if  defined(LINUX)
   int      err = mkdir(dirname, (0777 & ~umask(0)) | 0300);
#else
   int      err = mkdir(dirname);
#endif    // LINUX

   if   (0==err)  err = chdir(dirname);
   if   (0==err)  return RemoteSendCommand("OK");

   AddToMemo("\n(a) No write permission in parent directory, or");
   AddToMemo("\n(b) File exists but is not a directory");

// sprintf(scratch,	    "DIR ERROR %u", 1);	// ???
   return RemoteSendCommand("DIR ERROR 1");	// "Cannot create dir"
			// (a) no write permission in parent directory
			// (b) file exists but is not a directory
}


   static void
NoWritePermission(void)  // Reports "No write permission"
{  AddToMemo("\nProbably no write permission in parent directory"); }


   static boolean
ServerDeleteFile(char *filename)	// Deletes a file on the server
{
   // 'int unlink(char*)' is the time-honoured Unix function for this
   //  int remove(char*)  is the MS-DOS name
   int err = remove(FormatUnixName(filename));
/*
   FILE *f = NULL;
   assign(&f,filename);
   erase(f);
   err = ioresult();
   if   (err) {
     setfattr(filename,dos_archive);   // #define dos_archive  _A_ARCH
     erase(f);
     err = ioresult();
   }
 */
   if   (0==err)  return RemoteSendCommand("OK");

   NoWritePermission();
   doserror = ENOENT /*errno*/;  // "Cannot remove file"
   return RemoteSendCommand("FILE ERROR 2");
}


   static boolean
ServerRemoveEmptyDirectory(char *dirname) // Deletes a directory on the server
{
   if (0==rmdir(FormatUnixName(dirname)))  return RemoteSendCommand("OK");

   NoWritePermission();
   doserror = ENOTDIR /*errno*/ ;  // "Cannot remove directory"
// sprintf(scratch, "DIR ERROR %u", doserror);
   return RemoteSendCommand("DIR ERROR 20");
}


   static boolean
ServerRenameFile(char *fromname, char *toname) // Renames a file on the server
{
   if (0==rename(fromname,toname))  return RemoteSendCommand("OK");
	// Can it also rename directories ? -- Yes.

   NoWritePermission();
   doserror = ENOENT /*errno*/ ;	// "Cannot rename file"
   return RemoteSendCommand("RENAME ERROR 2");
}


   static boolean
ServerSignalConnectOK(void)		// Transmits "OK A"
   { return RemoteSendCommand("OK A"); }


   static long
GetRestartAtRemote(long restartAt)	// Gets new 'restartAtRemote'
{
   // Used in 'textMode' only:
   // Maps (Linux) server byte count to (Windows) client byte count

   long result    = 0;			// = remotedataTotal
   long dataTotal = 0;
   long timehelp;
   long startTimeSeconds = GetEpochTime();

   while (dataTotal++ < restartAt  &&  EOF != ci  &&  EOF != GETci()) {
     ch = (char)ci;
     if   (ch == '\n')  result++;	// One more per \r\n
     else MapUTF8();
     result++;

     if  (0== (dataTotal & 0x7FFF)  &&	// Send WAIT after every 32 KB
	  startTimeSeconds < (timehelp = GetEpochTime())) {
	  startTimeSeconds =  timehelp;
	  if (! RemoteSendCommand("WAIT"))  { result= -1;  break; }
     }
   }

   return result;
}


   static boolean		// -- one reference only --
WaitForStatusBit(void)		// Waits for the wake-up signal
{				// Result:  true:success   false:failure
   do {
     if (0==PortStatus())  return true;	// Call to driver on NT
    _sleep(50L);			// 1/20 sec  pause
   } while (!(abortClicked = kbhit()));

   return false;
}


#define Starts_with(s,cmd)  (!strncmp(s,cmd,sizeof(cmd)-1)) // Tests string


   static void
DoServer(void)		// Accepts, and responds to, commands from client
{
   static char renamedFrom[310];	// Old file name for renaming
   static char     recName[310];	// File name for (A)PUT and SIZE
   static char		 s[256];	// Command string

   long   restartAtRemote = 0;
   long   restartAt =  0;
   long   recTime   = -1;
   int    recAttr   = -1;
   char   recStatus =  0;		// 0=Before, 1=DoneOK, -1=Error
   boolean  textPut = false;

   *recName     = '\0';
   *renamedFrom = '\0';

   SendDisableInterrupts();
   SetPortReady();
   abortClicked = false;

   while (! abortClicked) {
     *s = '\0';
     if (WaitForStatusBit()) {		// Waiting for wake-up signal
       recStatus = ReceiveCommandString(s);
       if   (!abortClicked) {
	 int hash = s[0]+s[2];
	 if (*s && ('T'+'M') != hash  &&  ('A'+'T') != hash)
			  { AddToMemo("\n"); AddToMemo(s); } // Protocol echo

	 switch (hash) {
	   case 'P'+'T':  if (!Starts_with(s,"PUT ")) goto fail;
			  strcpy(recName, s+4);			  // TO server
			  recTime  = -1;
			  recAttr  = -1;
			  textPut  = false;
			  break;

	   case 'A'+'U':  if (!Starts_with(s,"APUT ")) goto fail;
			  strcpy(recName, s+5);			  // TO server
			  recTime  = -1;
			  recAttr  = -1;
			  textPut  = true;
			  DOSmode  =  0;	// Character set now: ANSI
			  break;

	   case 'T'+'M':  if (!Starts_with(s,"TIME ")) goto fail;
			  recTime  = StrToInt(s+5);		  // TO server
			  break;

	   case 'A'+'T':  if (!Starts_with(s,"ATTR ")) goto fail;
			  recAttr  = StrToInt(s+5);		  // TO server
			  break;

	   case 'S'+'Z':  if (!Starts_with(s,"SIZE ")) goto fail; // TO server
			  if (0==ReceiveData(recName,StrToInt(s+5), recTime,
				recAttr,restartAt,restartAtRemote,textPut))
			       { AddToMemoOkay()	      ;recStatus=  1; }
			  else { AddToMemo("\nUPLOAD FAILED") ;recStatus= -1; }
			  textPut = false;
			  abortClicked = false;
			 *recName   ='\0';
			  recTime   = -1;
			  recAttr   = -1;
			  restartAt =  0;
			  restartAtRemote = 0;  // (ptp)
			  break;

	   case 'R'+'S':  if (!Starts_with(s,"REST ")) goto fail;
			  if ('?' != s[5]) {		 // "RESTART at nnn"
			      restartAt = StrToInt(s+5); // FROM+TO server
			      if (restartAt== -1)  restartAt=0;
			      break;
			  }
			  if (*recName) { // Reading (checking what's on disk)
			    doserror = 0;		 // TO server
			    ci       = 0;
			    if (NULL !=(fi=fopen(FormatUnixName(recName),
						(textPut?"r":"rb"))))  {
			      restartAt=  fstat(fileno(fi),&fprops) ?0 :
					 (fprops.st_size-4096) & (~0x3FF);
						// Clear bottom 10 bits
			      if  (restartAt < 0)   restartAt = 0;
			      restartAtRemote = (textPut &&  restartAt)
					? GetRestartAtRemote(restartAt)
					: restartAt;
			      fclose(fi);  fi=NULL;
			      if (0 <= restartAt) {
				sprintf(scratch, "REST=%lu", restartAtRemote);
				(void) RemoteSendCommand(scratch);
				break;
			      } else {
				restartAtRemote = 0;
				restartAt       = 0;
			      }
			    }
			  }
			  doserror = ENOENT /*errno*/ ;
			  (void) RemoteSendCommand("REST ERROR 2");
			  break;

	   case 'G'+'T':  if (!Starts_with(s,"GET ")) goto fail; // FROM server
			  if   (ServerUploadFile(s+4, restartAt,false))
			       { AddToMemoOkay()	       ;recStatus=  1;}
			  else { AddToMemo("\nDOWNLOAD FAILED");recStatus= -1;}
			  abortClicked    = false;
			  restartAt       = 0;
			  restartAtRemote = 0;
			  break;

	   case 'A'+'E':  if (!Starts_with(s,"AGET ")) goto fail; // FROM server
			  if   (ServerUploadFile(s+5, restartAt,true))
			       { AddToMemoOkay()	       ;recStatus=  1;}
			  else { AddToMemo("\nDOWNLOAD FAILED");recStatus= -1;}
			  abortClicked    = false;
			  restartAt       = 0;
			  restartAtRemote = 0;
			  break;

	   case 'C'+'N':  if (!Starts_with(s,"CONNECT")) goto fail;
			  if (ServerSignalConnectOK()) {
				 AddToMemo("ION ESTABLISHED\t\t\t\tOK");
				 recStatus =  1;
			  } else recStatus = -1;
			  break;

	   case 'L'+'S':  if   (!Starts_with(s,"LIST ")) goto fail;
			  textPut = false;
			  if   (ServerListDirectory(s+5))
			       { AddToMemoOkay()	     ;recStatus =  1; }
			  else { AddToMemo("\nLIST FAILED")  ;recStatus = -1; }
			  abortClicked = false;
			  break;

	   case 'D'+'L':  if (!Starts_with(s,"DEL ")) goto fail;
			  recStatus = ServerDeleteFile(s+4) ?1 :-1;
			  break;

	   case 'M'+' ':  if (!Starts_with(s,"MD ")) goto fail;
			  recStatus = ServerCreateDirectory(
					FormatUnixName(s+3)) ?1 :-1;
			  break;

	   case 'R'+' ':  if (!Starts_with(s,"RD ")) goto fail;
			  recStatus = ServerRemoveEmptyDirectory(s+3) ?1 :-1;
			  break;

	   case 'R'+'F':  if (!Starts_with(s,"RNFR ")) goto fail;
			  strcpy(renamedFrom, FormatUnixName(s+5));
			  break;

	   case 'R'+'T':  if (!Starts_with(s,"RNTO ")) goto fail;
			  recStatus = (*renamedFrom  &&
			    ServerRenameFile(renamedFrom, FormatUnixName(s+5)))
				      ?1 :-1;
			  break;

fail:      default :	  // Unknown command
			  AddToMemo("\nStill trying...");
			  if (*s) { printf(" [%s",s); AddToMemo("]"); }
			  SignalConnectionError();
			  SetPortReady();	// Clear error condition
			  abortClicked = kbhit();
			  break;
	 }		  // switch (hash) ...
       }		  // if (! abortClicked) ...
       *s = '\0';
     }			  // if (WaitForStatusBit()) ...
     if (recStatus== -1)  SetPortReady();
   }			  // while (! abortClicked) ...
}  // DoServer


   static int
FindPort(void)				// Finds parallel port automatically
{
   int  i, k;
   abortClicked = false;
   for  (k=150  ;0 <= --k  &&  !abortClicked ;)  // Trying for 45 seconds
   for  (i=4    ;0 <  --i  ;)		// Examining the 3 ports in turn
     if ((portaddr = GetPortAddress(i)))// (Note: it seems none ever fails)
     {
	 portaddr1 = portaddr+1;	// Read port (needed!)
	 SendDisableInterrupts();
	 SetPortReady();	// Now checking for input from Total Commander
	 if (WaitForBitFlip(false,2))	{ portnr[0]= '0'+i;  return portaddr; }
      // ioperm(portaddr,3,0);		// ioperm(...,0) is LETHAL!  DONT USE!
	 abortClicked = kbhit();
     }
 //printf("\nFailed! -- no response on any port.\n\n");  exit(1);
   return 0;
}


   int
main(int argc, char **argv)		// Main program
{
   int eUID;				// File access UID

   ftime (&time_a);			// Initialising time measurement
// printf("tz=%.1f\n", time_a.timezone/60.);

   printf("\nTC_SERV 1.0 -- Parallel port server for Total Commander.\n\n");
   printf("Usage:  ./tc_serv <portnr>     with <portnr> = 1|2|3 or empty\n");
   printf("This program can be aborted by pressing Ctrl/C\n");
   printf("Copyright 2000-2005 by Christian Ghisler, C. Ghisler & Co.\n");
   printf("Copyright 2004-2005 by Paul Theo Pilgram\n");
   printf("==========================================================\n");

#if  defined(LINUX)

  // Capturing segmentation faults only when we are not debugging...
#if !defined(DEBUG)
   (void) signal(SIGSEGV ,SignalHandler);
#endif

   (void) signal(SIGHUP  ,SignalHandler);
   (void) signal(SIGINT  ,SignalHandler); // Ctrl-C
   (void) signal(SIGTERM ,SignalHandler); // SW termination signal from kill
 //(void) signal(SIGQUIT ,SignalHandler); // -- ?
 //(void) signal(SIGIO   ,SignalHandler); // Any keyboard input
///(void) signal(SIGBREAK,SignalHandler); // Ctrl-Break sequence (U:undefined)
#endif					  // SI_SIGIO defined in bits/siginfo.h

   // ______ Initialising and starting the server: ____________________	_

    uid =  getuid();		//	    User UID
   eUID = geteuid();		// tc_serv owner UID (due to sticky bit)
   // The 'root sticky bit' gives the user root rights,
   // the eUID determines hardware and file access rights.
   // 'root' has UID 0
// printf("Current effective user ID: %u;  ID of calling user: %u\n",eUID,uid);

   if (0==uid || eUID) {	// If running as root or without sticky bit:
     printf("\nWARNING:\n   You can run this program as a normal user,\n");
     printf("   but it must have the \"root sticky bit\" set.\n");
     printf("   This is achieved with the following commands:\n");
     printf("\tsu\n\tchown root tc_serv\n\tchmod  +s  tc_serv\n\texit\n\n");
   }

   if (1<argc) portnr[0]=argv[1][0];	// Gets checked in GetPortAddress
   printf("Allocating port LPT%s:", portnr);  fflush(stdout);
   portaddr  = portnr[0]  ?GetPortAddress(portnr[0]-'0')  :FindPort();
   portaddr1 = portaddr+1;		// The read port

   if (0 == portaddr) { printf(
      "\nFailed! -- no response on any port.\n\n")		   ;return 1; }
 //   "\nFailed! -- wrong port or no port access permission.\n\n") ;return 1; }

   if (uid != eUID)			// If eUID=root and UID=user:
       seteuid(uid);			// Reducing eUID (root!) to user UID
       // Normal rights apply in all subsequent file accesses/creations.

   // ______ Running as server: _______________________________________	_

   abortClicked  = false;
   blocktransfer = false;

   printf(
   "\t\t\t\tOK\nAccepting connections through port LPT%s, port addr 0x%03X\n"
	 ,portnr, portaddr);

   DoServer();			//   <--   THIS IS WHERE IT'S ALL HAPPENING!

   AddToMemo("\nDISCONNECT\t\t\t\t\tOK\n");
   return 0;
}  // main


// ____________________________________________________________________	_


/* Comment from P T Pilgram:
   Our computer administrators find it hard to agree to any use
   of "sticky bit" programs on our networked Linux computers.

   Reply from Christian Ghisler:
   Give them the source code and show them that, after acquiring access to
   the parallel port, the program immediately surrenders the root privileges.
   That should allay their fears.

   An alternative (which might require reprogramming) to the sticky bit would
   be through granting parport access to this user or a particular group.
   ______________________________

Notes to anybody who wants to modify this program:

Character sets, UTF8 :
 Please read the comments, above, concerning UTF8 and the Euro symbol

Byte order of multibyte data:

 Care has been taken to achieve that this program runs on almost any
 architecture, although there was little chance for testing this.
 Computer inventors had diverging ideas about how to arrange things
 in their new designs.  On some machines, the byte addresses in 32-bit
 integers increase right to left ("little-endian"), on other machines
 they increase left to right ("big-endian"); the number of possible
 combinations is 4! = 24, and I fear each of them has been built.

 You can detect which endian your machine is by running this program:

 main()
 { long L=0x34333231;  printf("#define __BYTE_ORDER %.4s\n",(char*)&L); }

 // Alternatively:
 // char S[]="Test";*(long*)S=0x34333231; printf("#define __BYTE_ORDER %s",S);

 If it prints 1234, your machine is little-endian (e.g. Intel_386)
 If it prints 4321, your machine is    big-endian (e.g. Sun or Motorola)
 If it prints 3412, your machine is probably a DEC PDP.

 Program TC_SERV and the transfer protocol stems from an MS-DOS program;
 little-endian order is used when transmitting numbers (blocknumbers,
 blocksizes, CRCs).  The program adapts this order where required.

 Most implementations care only about little vs. big endian.
 These two can be told apart like this:
   int isLittleEndian= 1; // And now grab left or rightmost byte:
       isLittleEndian= *((char*) &isLittleEndian);
 */
